(ns cljs_ajax_battleship.fleet
  (:require [cljs_ajax_battleship.board :as board]
            [cljs_ajax_battleship.serialization :as serialization]
            [clojure.browser.repl :as repl]
            [goog.uri :as uri]
            [goog.uri.QueryData :as qd])
  (:use [jayq.core :only [$]]))

(def original-pos (atom {}))
(def last-pos (atom {}))
(def ship-map (atom {}))

(def ship-types [ "carrier" "battleship" "cruiser" "submarine" "destroyer" ])
(def ship-sizes (zipmap ship-types [5 4 3 3 2]))

(defn deploy-td-coll []
  (filter (fn [x] (> (count (.-id x)) 0))
          (js->clj ($ "#deploy td"))))

(defn contains-pt? [elem {x :x y :y}]
  (let [jqo      ($ elem)
        position (.position jqo)
        top      (.-top position)
        left     (.-left position)
        height   (.height jqo)
        width    (.width jqo)
        bottom   (+ top height)
        right    (+ left width)]
    (and (>= x left)
         (<= x right)
         (>= y top)
         (<= y bottom))))

(defn singleton-string-seq [a-string]
  (map str (seq a-string)))

;; (set (map str (seq a-string))) returns same under ClojureScript and Clojure
(defn singleton-string-set [a-string]
  (set (map str (seq a-string))))

(defn number-set []
  (singleton-string-set "0123456789"))

(defn coord-from-table-cell-id [id-str]
  (let [coord-chars    (singleton-string-set "ABCDEFGHIJ0123456789")
        is-coord-char? (fn [c] (contains? coord-chars c))]
    (apply str (filter is-coord-char? id-str))))

(defn coord-of-elem [elem]
  (coord-from-table-cell-id (aget elem "id")))

(defn midpoint [elem]
  (let [jqo      ($ elem)
        position (.position jqo)
        top      (.-top position)
        left     (.-left position)
        height   (.height jqo)
        width    (.width jqo)]
    {:x (+ left (/ width 2))
     :y (+ top (/ height 2))}))

(defn contains-midpoint? [elem1 elem2]
  (contains-pt? elem1 (midpoint elem2)))

(defn covered-coords [ship]
  (let [ship-elem      ($ ship)
        covers-target? (fn [elem] (contains-midpoint? ship-elem elem))
        covered        (filter covers-target? (deploy-td-coll))
        covered-ids    (map (fn [x] (coord-of-elem x)) covered)]
    (map (fn [x] (coord-of-elem x)) covered)))

(defn ship-type [id-str]
  (let [not-number? (fn [c] (not (contains? (number-set) c)))]
    (apply str (filter not-number? id-str))))

(defn counterpart-id [id-str]
  (let [suffix      (apply str (filter (number-set) (seq id-str)))
        inverted    (mod (+ 1 (js/parseInt suffix)) 2)]
    (str (ship-type id-str) inverted)))

(defn counterpart-ship [ship]
  (let [ship-id (.-id ship)]
    (counterpart-id ship-id)))

(defn return-to-origin [ship-id]
  (let [origin (get (deref original-pos) ship-id)]
     (.animate ($ (str "#" ship-id)) (apply js-obj origin))))

(defn return-to-last-pos [ship-id]
  (let [origin (get (deref last-pos) ship-id)]
     (.animate ($ (str "#" ship-id)) (apply js-obj origin))))

(defn get-current-ship-positions []
  (apply merge
         (for [dragthing ($ ".dragthing")]
           (let [position  (.position ($ dragthing))
                 x         (.-left position)
                 y         (.-top position)
                 point     ["left" x "top" y]]
             {(.-id dragthing) point}))))

(defn calc-ship-map []
  (apply merge
         (flatten 
          (for [dthing ($ ".dragthing")]
            (let [coords   (covered-coords dthing)
                  id-str   (.attr ($ dthing) "id")]
              (map (fn [c0] {c0 (ship-type id-str)}) coords))))))

(defn write-ship-map [ship-map]
  (.val ($ "#shipdata") (str ship-map)))

(defn start-ship-drag [event ui]
  (let [mytarget     (.-target event)
        my-id        (.-id mytarget)
        other-id     (counterpart-id my-id)
        other-ship   ($ (str "#" other-id))
        other-coords (covered-coords other-ship)]
    (do
      (apply swap! (concat [ship-map dissoc] other-coords))
      (return-to-origin (counterpart-id my-id)))))

(defn drag-ship [event ui] nil)

(defn stop-ship-drag [event ui]
  (let [mytarget        (.-target event)
        my-id           (.-id mytarget)
        coords          (covered-coords mytarget)
        is-empty-coord? (fn [c0] (not (contains? (deref ship-map) c0)))]
    (do
      (if (not (every? is-empty-coord? coords))
        (do
          (js/console.log (str "was not empty:" coords))
          (return-to-last-pos my-id))
        (do (swap! last-pos get-current-ship-positions) 
            (swap! ship-map calc-ship-map)
            (write-ship-map (deref ship-map)))))))

(defn init-ship-marker! [ship]
  (.draggable ship
              (js-obj "grid"  (array 28 28)
                      "start" start-ship-drag
                      "stop"  stop-ship-drag)))

(defn encode-query-string [hash-map]
  )

(defn start-game! []
  ; (optionally) validate board
  ; Serialize current board status
  ; Create computer board
  ; Submit to route
  (let [player (board/create-player {} (calc-ship-map))
        computer (board/create-player {} (board/rand-ships))]
    (get "/startgame" (encode-query-string {:p1 player
                                            :p2 computer}))))

(defn ^:export init []
  (do
    (swap! original-pos get-current-ship-positions)
    (doall (for [dthing ($ ".dragthing")]
             (init-ship-marker! ($ dthing))))
    (js/console.log "cljs initialized")))

;; (in-ns 'cljs_ajax_battleship.fleet)

(defn ^:export nrepl [] 
  (repl/connect "http://localhost:9000/repl"))

(set! (.-nrepl js/window) nrepl)
